# -×- coding: utf-8 -*-

import logging
import time

from threading import Thread
from multiprocessing import RLock

class HeartBeatLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        heart_beat = self.extra['heart_beat']  # type: ConsumerHeatBeat
        _id = '/'.join([
            heart_beat.log_client.mproject, heart_beat.log_client.mlogstore,
            heart_beat.log_client.mconsumer_group,
            heart_beat.log_client.mconsumer
        ])
        return "[{0}]{1}".format(_id, msg), kwargs


class ConsumerHeatBeat(Thread):

    def __init__(self, log_client, heartbeat_interval, consumer_group_time_out):
        super(ConsumerHeatBeat, self).__init__()
        self.log_client = log_client
        self.heartbeat_interval = heartbeat_interval
        self.mheld_shards = []
        self.mheart_shards = []
        self.shut_down_flag = False
        self.lock = RLock()
        self.last_hearbeat_successed_unixtime = time.time()
        self.consumer_group_time_out = consumer_group_time_out
        self.logger = HeartBeatLoggerAdapter(
            logging.getLogger(__name__), {"heart_beat": self})

    def run(self):
        self.logger.info('heart beat start')
        while not self.shut_down_flag:
            try:
                response_shards = []
                last_heatbeat_time = time.time()

                if self.log_client.heartbeat(self.mheart_shards, response_shards):
                    self.last_hearbeat_successed_unixtime = time.time()
                    self.logger.debug('heart beat result: %s get: %s',
                                      self.mheart_shards, response_shards)
                    if self.mheart_shards != response_shards:
                        current_set, response_set = set(
                            self.mheart_shards), set(response_shards)
                        add_set = response_set - current_set
                        remove_set = current_set - response_set
                        if any([add_set, remove_set]):
                            self.logger.info(
                                "shard reorganize, adding: %s, removing: %s",
                                add_set, remove_set)
                else:
                    if time.time() - self.last_hearbeat_successed_unixtime > \
                            (self.consumer_group_time_out + self.heartbeat_interval):
                        response_shards = []
                        self.logger.info(
                            "Heart beat timeout, automatic reset consumer held shards")
                    else:
                        with self.lock:
                            response_shards = self.mheld_shards
                            self.logger.info(
                                "Heart beat failed, Keep the held shards unchanged")

                with self.lock:
                    self.mheart_shards = list(set(self.mheart_shards + response_shards))
                    self.mheld_shards = response_shards

                # default sleep for 2s from "LogHubConfig"
                time_to_sleep = self.heartbeat_interval - (time.time() - last_heatbeat_time)
                while time_to_sleep > 0 and not self.shut_down_flag:
                    time.sleep(min(time_to_sleep, 1))
                    time_to_sleep = self.heartbeat_interval - (time.time() - last_heatbeat_time)
            except Exception as e:
                self.logger.warning("fail to heat beat", e)

        self.logger.info('heart beat exit')

    def get_held_shards(self):
        """
        must copy to prevent race condition in multi-threads
        :return:
        """
        return self.mheld_shards[:]

    def shutdown(self):
        self.logger.info('try to stop heart beat')
        self.shut_down_flag = True

    def remove_heart_shard(self, shard):
        self.logger.info('try to remove shard "{0}", current shard: {1}'.format(shard, self.mheld_shards))
        with self.lock:
            if shard in self.mheld_shards:
                self.mheld_shards.remove(shard)
            if shard in self.mheart_shards:
                self.mheart_shards.remove(shard)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

__all__ = ['CreateSqlInstanceResponse', 'UpdateSqlInstanceResponse', 'ListSqlInstanceResponse']

from .logresponse import LogResponse


class CreateSqlInstanceResponse(LogResponse):
    """ The response of the create_sql_instance API from log.

    :type header: dict
    :param header: CreateSqlInstanceResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateSqlInstanceResponse:')
        print('request_id:', self.get_request_id())
        print('headers:', self.get_all_headers())
        print('response:', self.get_body())


class UpdateSqlInstanceResponse(LogResponse):
    """ The response of the update_sql_instance API from log.

    :type header: dict
    :param header: UpdateSqlInstanceResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateSqlInstanceResponse:')
        print('request_id:', self.get_request_id())
        print('headers:', self.get_all_headers())
        print('response:', self.get_body())


class ListSqlInstanceResponse(LogResponse):
    """ The response of the list_sql_instance API from log.

    :type header: dict
    :param header: ListSqlInstanceResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('ListSqlInstanceResponse:')
        print('request_id:', self.get_request_id())
        print('headers:', self.get_all_headers())
        print('response:', self.get_body())


import logging
from .logclient import LogClient
from .logitem import LogItem
from .putlogsrequest import PutLogsRequest
from threading import Thread
import atexit
from time import time, sleep
from enum import Enum
from .version import LOGGING_HANDLER_USER_AGENT
import six
try:
    from collections.abc import Callable	
except ImportError:	
    from collections import Callable
import six

if six.PY2:
    from Queue import Empty, Full, Queue
else:
    from queue import Empty, Full, Queue

import json
import re


class LogFields(Enum):
    """fields used to upload automatically
    Possible fields:
    record_name, level, func_name, module,
    file_path, line_no, process_id,
    process_name, thread_id, thread_name
    """
    record_name = 'name'
    level = 'levelname'
    file_name = 'filename'
    func_name = 'funcName'
    module = 'module'
    file_path = 'pathname'
    line_no = 'lineno'
    process_id = 'process'
    process_name = 'processName'
    thread_id = 'thread'
    thread_name = 'threadName'

    level_no = 'levelno'
    asc_time = 'asctime'
    created_timestamp = 'created'
    micro_second = 'msecs'
    relative_created = 'relativeCreated'


DEFAULT_RECORD_LOG_FIELDS = set((LogFields.record_name, LogFields.level,
                                 LogFields.func_name, LogFields.module,
                                 LogFields.file_path, LogFields.line_no,
                                 LogFields.process_id, LogFields.process_name,
                                 LogFields.thread_id, LogFields.thread_name))

BLACK_FIELD_LIST = set(['exc_info', 'exc_text', 'stack_info', 'msg', 'args', 'message'])

BUILTIN_LOG_FIELDS_NAMES = set(x for x in dir(LogFields) if not x.startswith('__'))
BUILTIN_LOG_FIELDS_NAMES.update(set(LogFields[x].value for x in BUILTIN_LOG_FIELDS_NAMES))
BUILTIN_LOG_FIELDS_NAMES.update(BLACK_FIELD_LIST)


class SimpleLogHandler(logging.Handler, object):
    """
    SimpleLogHandler, blocked sending any logs, just for simple test purpose

    :param end_point: log service endpoint

    :param access_key_id: access key id

    :param access_key: access key

    :param project: project name

    :param log_store: logstore name

    :param topic: topic, by default is empty

    :param fields: list of LogFields or list of names of LogFields, default is LogFields.record_name, LogFields.level, LogFields.func_name, LogFields.module, LogFields.file_path, LogFields.line_no, LogFields.process_id, LogFields.process_name, LogFields.thread_id, LogFields.thread_name, you could also just use he string name like 'thread_name', it's also possible customize extra fields in this list by disable extra fields and put white list here.

    :param buildin_fields_prefix: prefix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param buildin_fields_suffix: suffix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param extract_json: if extract json automatically, default is False

    :param extract_json_drop_message: if drop message fields if it's JSON and extract_json is True, default is False

    :param extract_json_prefix: prefix of fields extracted from json when extract_json is True. default is ""

    :param extract_json_suffix: suffix of fields extracted from json when extract_json is True. default is empty

    :param extract_kv: if extract kv like k1=v1 k2="v 2" automatically, default is False

    :param extract_kv_drop_message: if drop message fields if it's kv and extract_kv is True, default is False

    :param extract_kv_prefix: prefix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_suffix: suffix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_sep: separator for KV case, defualt is '=', e.g. k1=v1

    :param extra: if show extra info, default True to show all. default is True. Note: the extra field will also be handled with buildin_fields_prefix/suffix

    :param kwargs: other parameters  passed to logging.Handler
    """

    def __init__(self, end_point, access_key_id, access_key, project, log_store, topic=None, fields=None,
                 buildin_fields_prefix=None, buildin_fields_suffix=None,
                 extract_json=None, extract_json_drop_message=None,
                 extract_json_prefix=None, extract_json_suffix=None,
                 extract_kv=None, extract_kv_drop_message=None,
                 extract_kv_prefix=None, extract_kv_suffix=None,
                 extract_kv_sep=None, extra=None,
                 **kwargs):
        logging.Handler.__init__(self, **kwargs)
        self.end_point = end_point
        self.access_key_id = access_key_id
        self.access_key = access_key
        self.project = project
        self.log_store = log_store
        self.client = None
        self.topic = topic
        self.fields = DEFAULT_RECORD_LOG_FIELDS if fields is None else set(fields)

        self.extract_json = False if extract_json is None else extract_json
        self.extract_json_prefix = "" if extract_json_prefix is None else extract_json_prefix
        self.extract_json_suffix = "" if extract_json_suffix is None else extract_json_suffix
        self.extract_json_drop_message = False if extract_json_drop_message is None else extract_json_drop_message
        self.buildin_fields_prefix = "" if buildin_fields_prefix is None else buildin_fields_prefix
        self.buildin_fields_suffix = "" if buildin_fields_suffix is None else buildin_fields_suffix

        self.extract_kv = False if extract_kv is None else extract_kv
        self.extract_kv_prefix = "" if extract_kv_prefix is None else extract_kv_prefix
        self.extract_kv_suffix = "" if extract_kv_suffix is None else extract_kv_suffix
        self.extract_kv_drop_message = False if extract_kv_drop_message is None else extract_kv_drop_message
        self.extract_kv_sep = "=" if extract_kv_sep is None else extract_kv_sep
        self.extract_kv_ptn = self._get_extract_kv_ptn()
        self.extra = True if extra is None else extra
        self._skip_message = False
        self._built_in_root_filed = ""
        self._log_tags = None
        self._source = None

    @property
    def skip_message(self):
        return self._skip_message

    @skip_message.setter
    def skip_message(self, value):
        self._skip_message = value

    @property
    def built_in_root_field(self):
        return self._built_in_root_filed

    @built_in_root_field.setter
    def built_in_root_field(self, value):
        self._built_in_root_filed = value

    @property
    def log_tags(self):
        return self._log_tags

    @log_tags.setter
    def log_tags(self, value):
        self._log_tags = value

    def set_topic(self, topic):
        self.topic = topic

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        self._source = value

    def create_client(self):
        self.client = LogClient(self.end_point, self.access_key_id, self.access_key)
        self.client.set_user_agent(LOGGING_HANDLER_USER_AGENT)

    def send(self, req):
        if self.client is None:
            self.create_client()
        return self.client.put_logs(req)

    def set_fields(self, fields):
        self.fields = fields

    @staticmethod
    def _n(v):
        if v is None:
            return ""

        if isinstance(v, (dict, list, tuple)):
            try:
                v = json.dumps(v)
            except Exception:
                pass
        elif six.PY2 and isinstance(v, six.text_type):
            v = v.encode('utf8', "ignore")
        elif six.PY3 and isinstance(v, six.binary_type):
            v = v.decode('utf8', "ignore")

        return str(v)

    def extract_dict(self, message):
        data = []
        if isinstance(message, dict):
            for k, v in six.iteritems(message):
                data.append(("{0}{1}{2}".format(self.extract_json_prefix, self._n(k),
                                                self.extract_json_suffix), self._n(v)))
        return data

    def _get_extract_kv_ptn(self):
        sep = self.extract_kv_sep
        p1 = u'(?!{0})([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)\\s*{0}\\s*([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)'
        p2 = u'(?!{0})([\u4e00-\u9fa5\u0800-\u4e00\\w\\.\\-]+)\\s*{0}\\s*"\s*([^"]+?)\s*"'
        ps = '|'.join([p1, p2]).format(sep)

        return re.compile(ps)

    def extract_kv_str(self, message):

        if isinstance(message, six.binary_type):
            message = message.decode('utf8', 'ignore')

        r = self.extract_kv_ptn.findall(message)

        data = []
        for k1, v1, k2, v2 in r:
            if k1:
                data.append(("{0}{1}{2}".format(self.extract_kv_prefix, self._n(k1),
                                                self.extract_kv_suffix), self._n(v1)))
            elif k2:
                data.append(("{0}{1}{2}".format(self.extract_kv_prefix, self._n(k2),
                                                self.extract_kv_suffix), self._n(v2)))

        return data

    def _add_record_fields(self, record, k, contents):
        data = self._get_record_fields(record, k)
        if data:
            contents.append(data)

    def _get_record_fields(self, record, k):
        v = getattr(record, k, None)
        if v is None or isinstance(v, Callable):
            return

        v = self._n(v)
        return "{0}{1}{2}".format(self.buildin_fields_prefix, k, self.buildin_fields_suffix), v

    def make_request(self, record):
        # add builtin fields
        built_root = {}
        contents = []
        message_field_name = "{0}message{1}".format(self.buildin_fields_prefix, self.buildin_fields_suffix)
        if isinstance(record.msg, dict) and self.extract_json:
            data = self.extract_dict(record.msg)
            contents.extend(data)

            if not self.extract_json_drop_message or not data:
                if not self._skip_message:
                    if self._built_in_root_filed:
                        built_root[message_field_name] = self.format(record)
                    else:
                        contents.append((message_field_name, self.format(record)))
        elif isinstance(record.msg, (six.text_type, six.binary_type)) and self.extract_kv:
            data = self.extract_kv_str(record.msg)
            contents.extend(data)

            if not self.extract_kv_drop_message or not data:  # if it's not KV
                if not self._skip_message:
                    if self._built_in_root_filed:
                        built_root[message_field_name] = self.format(record)
                    else:
                        contents.append((message_field_name, self.format(record)))
        elif not self._skip_message:
            if self._built_in_root_filed:
                built_root[message_field_name] = self.format(record)
            else:
                contents = [(message_field_name, self.format(record))]

        for x in self.fields:
            k = x
            if isinstance(x, LogFields):
                k = x.name
                x = x.value
            elif isinstance(x, (six.binary_type, six.text_type)):
                if x in BLACK_FIELD_LIST:
                    continue  # by pass for those reserved fields. make no sense to render them

                if x in BUILTIN_LOG_FIELDS_NAMES:
                    k = LogFields[x].name
                    x = LogFields[x].value
                elif self.extra:  # will handle it later
                    continue
            if not self._built_in_root_filed:
                self._add_record_fields(record, x, contents)
            else:
                data = self._get_record_fields(record, x)
                if data:
                    built_root[data[0]] = data[1]

        if self._built_in_root_filed and built_root:
            contents.append((self._n(self._built_in_root_filed), json.dumps(built_root)))

        # handle extra
        if self.extra:
            for x in dir(record):
                if not x.startswith('__') and not x in BUILTIN_LOG_FIELDS_NAMES:
                    self._add_record_fields(record, x, contents)

        item = LogItem(contents=contents, timestamp=record.created)

        return PutLogsRequest(self.project, self.log_store, self.topic, source=self._source, logitems=[item, ], logtags=self._log_tags)

    def emit(self, record):
        try:
            req = self.make_request(record)
            self.send(req)
        except Exception as e:
            self.handleError(record)


class QueuedLogHandler(SimpleLogHandler):
    """
    Queued Log Handler, tuned async log handler.

    :param end_point: log service endpoint

    :param access_key_id: access key id

    :param access_key: access key

    :param project: project name

    :param log_store: logstore name

    :param topic: topic, default is empty

    :param fields: list of LogFields, default is LogFields.record_name, LogFields.level, LogFields.func_name, LogFields.module, LogFields.file_path, LogFields.line_no, LogFields.process_id, LogFields.process_name, LogFields.thread_id, LogFields.thread_name

    :param queue_size: queue size, default is 40960 logs, about 10MB ~ 40MB

    :param put_wait: maximum delay to send the logs, by default 2 seconds and wait double time for when Queue is full.

    :param close_wait: when program exit, it will try to send all logs in queue in this timeperiod, by default 5 seconds

    :param batch_size: merge this cound of logs and send them batch, by default min(1024, queue_size)

    :param buildin_fields_prefix: prefix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param buildin_fields_suffix: suffix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param extract_json: if extract json automatically, default is False

    :param extract_json_drop_message: if drop message fields if it's JSON and extract_json is True, default is False

    :param extract_json_prefix: prefix of fields extracted from json when extract_json is True. default is ""

    :param extract_json_suffix: suffix of fields extracted from json when extract_json is True. default is empty

    :param extract_kv: if extract kv like k1=v1 k2="v 2" automatically, default is False

    :param extract_kv_drop_message: if drop message fields if it's kv and extract_kv is True, default is False

    :param extract_kv_prefix: prefix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_suffix: suffix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_sep: separator for KV case, defualt is '=', e.g. k1=v1

    :param extra: if show extra info, default True to show all. default is True

    :param kwargs: other parameters  passed to logging.Handler
    """

    def __init__(self, end_point, access_key_id, access_key, project, log_store, topic=None, fields=None,
                 queue_size=None, put_wait=None, close_wait=None, batch_size=None,
                 buildin_fields_prefix=None, buildin_fields_suffix=None,
                 extract_json=None, extract_json_drop_message=None,
                 extract_json_prefix=None, extract_json_suffix=None,
                 extract_kv=None, extract_kv_drop_message=None,
                 extract_kv_prefix=None, extract_kv_suffix=None,
                 extract_kv_sep=None,
                 extra=None,
                 **kwargs):
        super(QueuedLogHandler, self).__init__(end_point, access_key_id, access_key, project, log_store,
                                               topic=topic, fields=fields,
                                               extract_json=extract_json,
                                               extract_json_drop_message=extract_json_drop_message,
                                               extract_json_prefix=extract_json_prefix,
                                               extract_json_suffix=extract_json_suffix,
                                               buildin_fields_prefix=buildin_fields_prefix,
                                               buildin_fields_suffix=buildin_fields_suffix,
                                               extract_kv=extract_kv,
                                               extract_kv_drop_message=extract_kv_drop_message,
                                               extract_kv_prefix=extract_kv_prefix,
                                               extract_kv_suffix=extract_kv_suffix,
                                               extract_kv_sep=extract_kv_sep,
                                               extra=extra,
                                               **kwargs)
        self.stop_flag = False
        self.stop_time = None
        self.put_wait = put_wait or 2  # default is 2 seconds
        self.close_wait = close_wait or 5  # default is 5 seconds
        self.queue_size = queue_size or 40960  # default is 40960, about 10MB ~ 40MB
        self.batch_size = min(batch_size or 1024, self.queue_size)  # default is 1024 items

        self.init_worker()

    def init_worker(self):
        self.worker = Thread(target=self._post)
        self.queue = Queue(self.queue_size)
        self.worker.setDaemon(True)

        self.worker.start()
        atexit.register(self.stop)

    def flush(self):
        self.stop()

    def stop(self):
        self.stop_time = time()
        self.stop_flag = True
        self.worker.join(timeout=self.close_wait + 1)

    def emit(self, record):
        req = self.make_request(record)
        req.__record__ = record
        try:
            self.queue.put(req, timeout=self.put_wait*2)
        except Full as ex:
            self.handleError(record)

    def _get_batch_requests(self, timeout=None):
        """try to get request as fast as possible, once empty and stop flag or time-out, just return Empty"""
        reqs = []
        s = time()
        while len(reqs) < self.batch_size and (time() - s) < timeout:
            try:
                req = self.queue.get(block=False)
                self.queue.task_done()

                reqs.append(req)
            except Empty as ex:
                if self.stop_flag:
                    break
                else:
                    sleep(0.1)

        if not reqs:
            raise Empty
        elif len(reqs) <= 1:
            return reqs[0]
        else:
            logitems = []
            req = reqs[0]
            for req in reqs:
                logitems.extend(req.get_log_items())

            ret = PutLogsRequest(self.project, self.log_store, req.topic, logitems=logitems, logtags=self._log_tags)
            ret.__record__ = req.__record__

            return ret

    def _post(self):
        while not self.stop_flag or (time() - self.stop_time) <= self.close_wait:
            try:
                req = self._get_batch_requests(timeout=self.put_wait)
            except Empty as ex:
                if self.stop_flag:
                    break
                else:
                    continue

            try:
                self.send(req)
            except Exception as ex:
                self.handleError(req.__record__)


class UwsgiQueuedLogHandler(QueuedLogHandler):
    """
    Queued Log Handler for Uwsgi, depends on library `uwsgidecorators`, need to deploy it separatedly.

    :param end_point: log service endpoint

    :param access_key_id: access key id

    :param access_key: access key

    :param project: project name

    :param log_store: logstore name

    :param topic: topic, default is empty

    :param fields: list of LogFields, default is LogFields.record_name, LogFields.level, LogFields.func_name, LogFields.module, LogFields.file_path, LogFields.line_no, LogFields.process_id, LogFields.process_name, LogFields.thread_id, LogFields.thread_name

    :param queue_size: queue size, default is 40960 logs, about 10MB ~ 40MB

    :param put_wait: maximum delay to send the logs, by default 2 seconds and wait double time for when Queue is full.

    :param close_wait: when program exit, it will try to send all logs in queue in this timeperiod, by default 2 seconds

    :param batch_size: merge this cound of logs and send them batch, by default min(1024, queue_size)

    :param buildin_fields_prefix: prefix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param buildin_fields_suffix: suffix of builtin fields, default is empty. suggest using "__" when extract json is True to prevent conflict.

    :param extract_json: if extract json automatically, default is False

    :param extract_json_drop_message: if drop message fields if it's JSON and extract_json is True, default is False

    :param extract_json_prefix: prefix of fields extracted from json when extract_json is True. default is ""

    :param extract_json_suffix: suffix of fields extracted from json when extract_json is True. default is empty

    :param extract_kv: if extract kv like k1=v1 k2="v 2" automatically, default is False

    :param extract_kv_drop_message: if drop message fields if it's kv and extract_kv is True, default is False

    :param extract_kv_prefix: prefix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_suffix: suffix of fields extracted from KV when extract_json is True. default is ""

    :param extract_kv_sep: separator for KV case, defualt is '=', e.g. k1=v1

    :param extra: if show extra info, default True to show all. default is True

    :param kwargs: other parameters  passed to logging.Handler
    """
    def __init__(self, *args, **kwargs):
        # change close_wait from default 5 to 2
        if len(args) >= 10:
            if args[9] is None:
                args = args[:9] + (2,) + args[10:]
        elif 'close_wait' in kwargs and kwargs['close_wait'] is None:
            kwargs['close_wait'] = 2

        super(UwsgiQueuedLogHandler, self).__init__(*args, **kwargs)

    def init_worker(self):
        self.queue = Queue(self.queue_size)

        from uwsgidecorators import postfork, thread
        self._post = postfork(thread(self._post))

    def stop(self):
        self.stop_time = time()
        self.stop_flag = True




# -×- coding: utf-8 -*-

import logging

from .exceptions import CheckPointException
from .exceptions import ClientWorkerException
from ..logexception import LogException
from ..version import USER_AGENT


class ConsumerClientLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        consumer_client = self.extra['consumer_client']  # type: ConsumerClient
        _id = '/'.join([
            consumer_client.mproject, consumer_client.mlogstore,
            consumer_client.mconsumer_group, consumer_client.mconsumer
        ])
        return "[{0}]{1}".format(_id, msg), kwargs


class ConsumerClient(object):
    def __init__(self, endpoint, access_key_id, access_key, project,
                 logstore, consumer_group, consumer, security_token=None, credentials_refresher=None):

        from .. import LogClient

        self.mclient = LogClient(endpoint, access_key_id, access_key, security_token)
        self.mclient.set_user_agent('%s-consumergroup-%s' % (USER_AGENT, consumer_group))
        if credentials_refresher is not None:
            self.mclient.set_credentials_auto_refresher(credentials_refresher)
        self.mproject = project
        self.mlogstore = logstore
        self.mconsumer_group = consumer_group
        self.mconsumer = consumer
        self.logger = ConsumerClientLoggerAdapter(
            logging.getLogger(__name__), {"consumer_client": self})

    def create_consumer_group(self, timeout, in_order):
        try:
            self.mclient.create_consumer_group(self.mproject, self.mlogstore, self.mconsumer_group,
                                               timeout, in_order)
        except LogException as e:
            # consumer group already exist
            if e.get_error_code() == 'ConsumerGroupAlreadyExist':

                try:
                    self.mclient.update_consumer_group(self.mproject, self.mlogstore, self.mconsumer_group,
                                                       timeout, in_order)

                except LogException as e1:
                    raise ClientWorkerException("error occour when update consumer group, errorCode: " +
                                                e1.get_error_code() + ", errorMessage: " + e1.get_error_message())

            else:
                raise ClientWorkerException('error occour when create consumer group, errorCode: '
                                            + e.get_error_code() + ", errorMessage: " + e.get_error_message())

    def get_consumer_group(self):
        for consumer_group in self.mclient.list_consumer_group(self.mproject, self.mlogstore).get_consumer_groups():
            if consumer_group.get_consumer_group_name() == self.mconsumer_group:
                return consumer_group

        return None

    def heartbeat(self, shards, responce=None):
        if responce is None:
            responce = []

        try:
            responce.extend(
                self.mclient.heart_beat(self.mproject, self.mlogstore,
                                        self.mconsumer_group, self.mconsumer, shards).get_shards())
            return True
        except LogException as e:
            self.logger.warning(e)

        return False

    def update_check_point(self, shard, consumer, check_point):
        self.mclient.update_check_point(self.mproject, self.mlogstore, self.mconsumer_group,
                                        shard, check_point, consumer)

    def get_check_point(self, shard):
        check_points = self.mclient.get_check_point(self.mproject, self.mlogstore, self.mconsumer_group, shard) \
            .get_consumer_group_check_points()

        if check_points is None or len(check_points) == 0:
            raise CheckPointException('fail to get shard check point')
        else:
            return check_points[0]

    def get_cursor(self, shard_id, start_time):
        return self.mclient.get_cursor(self.mproject, self.mlogstore, shard_id, start_time).get_cursor()

    def get_begin_cursor(self, shard_id):
        return self.mclient.get_begin_cursor(self.mproject, self.mlogstore, shard_id).get_cursor()

    def get_end_cursor(self, shard_id):
        return self.mclient.get_end_cursor(self.mproject, self.mlogstore, shard_id).get_cursor()

    def pull_logs(self, shard_id, cursor, end_cursor=None, count=1000):
        return self.mclient.pull_logs(self.mproject, self.mlogstore, shard_id, cursor, count, compress=True, end_cursor=end_cursor)



from .logresponse import LogResponse

__all__ = ['CreateEntityResponse', 'UpdateEntityResponse', 'DeleteEntityResponse', 'GetEntityResponse',
           'ListEntityResponse']


class CreateEntityResponse(LogResponse):
    pass


class UpdateEntityResponse(LogResponse):
    pass


class DeleteEntityResponse(LogResponse):
    pass


class GetEntityResponse(LogResponse):
    def get_entity(self):
        """
        Get entity itself
        :return:
        """
        return self.body

    def log_print(self):
        print('header: ', self.headers)
        print('body: ', self.body)


class ListEntityResponse(LogResponse):
    def __init__(self, header, resp, resource_name=None, entities_key=None):
        LogResponse.__init__(self, header, resp)
        self.count = resp['count']
        self._total = resp['total']
        backup_resource_name = ''
        if not resource_name:
            for x in resp:
                if x not in ('count', 'total'):
                    if x.endswith('s'):
                        resource_name = x
                        break
                    else:
                        backup_resource_name = x
            else:
                resource_name = backup_resource_name

        self.resource_name = resource_name
        if entities_key is None:
            entities_key = resource_name
        self.entities = resp.get(entities_key, [])

        if self.resource_name:
            setattr(self, 'get_' + resource_name, self.get_entities)

    def get_entities(self):
        """
        Get entities
        :return:
        """
        return self.entities

    def get_count(self):
        """ Get total count of entities

        :return: int, the number of total entities from the response
        """
        return self.count

    def get_total(self):
        """
        Get total count
        :return:
        """
        return self._total

    @property
    def total(self):
        return self._total

    def log_print(self):
        print('ListResponse for {0}:'.format(self.resource_name))
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('total:', self._total)
        print(self.resource_name, self.entities)

    def merge(self, response):
        if not isinstance(response, ListEntityResponse):
            raise ValueError("passed response is not a ListEntityResponse: " + str(type(response)))

        self.count += response.get_count()
        self._total = response.get_total()  # use the latest total count
        self.entities.extend(response.get_entities())

        # update body
        self.body = {
            'count': self.count,
            'total': self._total,
            self.resource_name: self.entities
        }

        return self




#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .util import Util
from .logresponse import LogResponse
from .logtail_config_detail import LogtailConfigGenerator

__all__ = ['CreateLogtailConfigResponse', 'DeleteLogtailConfigResponse',
           'GetLogtailConfigResponse', 'UpdateLogtailConfigResponse',
           'ListLogtailConfigResponse']


class CreateLogtailConfigResponse(LogResponse):
    """ The response of the create_logtail_config API from log.
    
    :type header: dict
    :param header: CreateLogtailConfigResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateLogtailConfigResponse:')
        print('headers:', self.get_all_headers())


class DeleteLogtailConfigResponse(LogResponse):
    """ The response of the delete_logtail_config API from log.
    
    :type header: dict
    :param header: DeleteLogtailConfigResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteLogtailConfigResponse:')
        print('headers:', self.get_all_headers())


class GetLogtailConfigResponse(LogResponse):
    """ The response of the get_logtail_config API from log.
    
    :type header: dict
    :param header: GetLogtailConfigResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.logtail_config = LogtailConfigGenerator.generate_config(resp)

    def log_print(self):
        print('GetLogtailConfigResponse:')
        print('headers:', self.get_all_headers())
        print('logtail_config:', self.logtail_config.to_json())


class UpdateLogtailConfigResponse(LogResponse):
    """ The response of the update_logtail_config API from log.
    
    :type header: dict
    :param header: UpdateLogtailConfigResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateLogtailConfigResponse:')
        print('headers:', self.get_all_headers())


class ListLogtailConfigResponse(LogResponse):
    """ The response of the list_logtail_config API from log.
    
    :type header: dict
    :param header: ListLogtailConfigResponse HTTP response header

    :type resp: dict
    :param resp: the HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total_count = int(resp["total"])
        self.logtail_configs = Util.convert_unicode_to_str(resp.get("configs", []))

    def get_configs(self):
        return self.logtail_configs

    def get_configs_count(self):
        return self.count

    def get_configs_total(self):
        return self.total_count

    def get_count(self):
        return self.count

    def get_total(self):
        return self.total_count

    @property
    def total(self):
        return self.total_count

    def log_print(self):
        print('ListLogtailConfigResponse:')
        print('headers:', self.get_all_headers())
        print('configs_count:', str(self.count))
        print('configs_total:', str(self.total_count))
        print('configs:', str(self.logtail_configs))

    def merge(self, response):
        if not isinstance(response, ListLogtailConfigResponse):
            raise ValueError("passed response is not a ListLogtailConfigResponse: " + str(type(response)))

        self.count += response.get_configs_count()
        self.total_count = response.get_configs_total() # use the latest total count
        self.logtail_configs.extend(response.get_configs())

        # update body
        self.body = {
            'count': self.count,
            'total': self.total_count,
            'configs': self.logtail_configs
        }

        return self



# -*- coding: utf-8 -*-

from enum import Enum


class CursorPosition(Enum):
    BEGIN_CURSOR = 'BEGIN_CURSOR'
    END_CURSOR = 'END_CURSOR'
    SPECIAL_TIMER_CURSOR = 'SPECIAL_TIMER_CURSOR'


class ConsumerStatus(Enum):
    INITIALIZING = 'INITIALIZING'
    PROCESSING = 'PROCESSING'
    SHUTTING_DOWN = 'SHUTTING_DOWN'
    SHUTDOWN_COMPLETE = 'SHUTDOWN_COMPLETE'


class LogHubConfig(object):

    def __init__(self, endpoint, access_key_id, access_key, project, logstore,
                 consumer_group_name, consumer_name,
                 cursor_position=None, heartbeat_interval=None, data_fetch_interval=None, in_order=False,
                 cursor_start_time=None, security_token=None, max_fetch_log_group_size=None, worker_pool_size=None, shared_executor=None,
                 cursor_end_time=None, credentials_refresher=None):
        """

        :param endpoint:
        :param access_key_id:
        :param access_key:
        :param project:
        :param logstore:
        :param consumer_group_name:
        :param consumer_name: suggest use format "{consumer_group_name}-{current_process_id}", give it different consumer name when you need to run this program in parallel
        :param cursor_position: This options is used for initialization, will be ignored once consumer group is created and each shard has beeen started to be consumed. normally, could ignore it.
        :param heartbeat_interval: default 20, once a client doesn't report to server * heartbeat_interval * 2 interval, server will consider it's offline and re-assign its task to another consumer. thus  don't set the heatbeat interval too small when the network badwidth or performance of consumtion is not so good.
        :param data_fetch_interval: default 2, don't configure it too small (<1s)
        :param in_order: default False, during consuption, when shard is splitted, if need to consume the newly splitted shard after its parent shard (read-only) is finished consumption or not. suggest keep it as False (don't care) until you have good reasion for it.
        :param cursor_start_time: Will be used when cursor_position when could be "begin", "end", "specific time format in ISO", it's log receiving time.
        :param security_token:
        :param max_fetch_log_group_size: default 1000, fetch size in each request, normally use default. maximum is 1000, could be lower. the lower the size the memory efficiency might be better.
        :param worker_pool_size: default 2. suggest keep the default size (2), use multiple process instead, when you need to have more concurrent processing, launch this consumer for mulitple times and give them different consuer name in same consumer group. will be ignored when shared_executor is passed.
        :param shared_executor: shared executor, if not None, worker_pool_size will be ignored
        :param cursor_end_time: cursor end time, default is None (never stop processing). could setting it as ISO time-format, when setting it as "end", it means process all logs received from start to the time when the consumer is started.
        """
        self.endpoint = endpoint
        self.accessKeyId = access_key_id
        self.accessKey = access_key
        self.project = project
        self.logstore = logstore
        self.consumer_group_name = consumer_group_name
        self.consumer_name = consumer_name
        self.cursor_position = cursor_position or CursorPosition.SPECIAL_TIMER_CURSOR
        self.heartbeat_interval = heartbeat_interval or 20
        self.data_fetch_interval = data_fetch_interval or 2
        self.in_order = in_order
        self.cursor_start_time = cursor_start_time or "begin"  # default to begin
        self.securityToken = security_token
        self.max_fetch_log_group_size = max_fetch_log_group_size or 1000
        self.worker_pool_size = worker_pool_size or 2
        self.shared_executor = shared_executor
        self.consumer_group_time_out = self.heartbeat_interval * 2
        self.cursor_end_time = cursor_end_time or None  # default to None
        self.credentials_refresher = credentials_refresher



import logging
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

import six
from ..trans_comp import REGEX
from .transform_base import transform_base
from ..etl_util import process_event, bind_event_fn, u

logger = logging.getLogger(__name__)

__all__ = ['transform']


def _dict_transform_fn(tr):
    def _real_transform(event):
        result = {}
        for k, v in six.iteritems(tr):
            if isinstance(v, Callable):
                v = v(event)

            if isinstance(v, (six.text_type, six.binary_type)):
                result[k] = v
            elif v is None:
                if k in result:
                    del result[k]
            else:
                logger.warning(u"unknown type of transform value for key:{0} value:{1}".format(k, v))
        event.update(result)
        return event

    return _real_transform


class transform(transform_base):
    def __init__(self, trans):
        trans = u(trans)

        if not isinstance(trans, list):
            self.trans = [trans]
        else:
            self.trans = trans

        self.transform_list = []
        for tr in self.trans:
            if isinstance(tr, Callable):
                self.transform_list.append(tr)
            elif isinstance(tr, (dict, )):
                self.transform_list.append(_dict_transform_fn(tr))
            elif isinstance(tr, tuple):
                if len(tr) < 2 or len(tr) > 3:
                    logger.warning(u"invalid transform config: {0}".format(tr))
                    continue

                inpt, config = tr[0], tr[1]
                if isinstance(config, (six.text_type, six.binary_type)):
                    self.transform_list.append(lambda e: REGEX(*tr[1:])(e, inpt))
                elif isinstance(config, Callable):
                    self.transform_list.append(bind_event_fn(config, inpt))
                else:
                    logger.warning(u"unknown transform config setting: {0}".format(config))
                    continue

    def __call__(self, event):
        return process_event(event, self.transform_list)



from .logresponse import LogResponse
from .resource_params import *

from .util import Util

__all__ = ['CreateResourceResponse', 'DeleteResourceResponse', 'GetResourceResponse', 'UpdateResourceResponse',
           "ListResourcesResponse", 'CreateRecordResponse', 'DeleteRecordResponse', "UpdateRecordResponse",
           "UpsertRecordResponse", "GetRecordResponse", "ListRecordResponse"
           ]


class CreateResourceResponse(LogResponse):
    """ The response of the create_resource API from log.

    :type resp: dict
    :param resp: CreateResourceResponse HTTP response body

    :type header: dict
    :param header: CreateResourceResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateResourceResponse:')
        print('headers:', self.get_all_headers())


class DeleteResourceResponse(LogResponse):
    """ The response of the delete_resource API from log.

    :type resp: dict
    :param resp: DeleteResourceResponse HTTP response body

    :type header: dict
    :param header: DeleteResourceResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteResourceResponse:')
        print('headers:', self.get_all_headers())


class ListResourcesResponse(LogResponse):
    """ The response of the list_resources API from log.

    :type header: dict
    :param header: ListResourcesResponse HTTP response header

    :type resp: dict
    :param resp: ListResourcesResponse HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total = int(resp["total"])
        self.resources = [Resource.from_dict(resource) for resource in Util.convert_unicode_to_str(resp.get("items", []))]

    def get_resources(self):
        """Get all the resources from the response

        :return: list, instances of Resource
        """
        return self.resources

    def get_count(self):
        """ Get count of resources from the response

        :return: int, the number of count resources from the response
        """
        return self.count

    def get_total(self):
        """Get total count of resources from the response

        :return: int, the number of total resources from the response
        """
        return self.total

    def log_print(self):
        print('ListResourcesResponse:')
        print('headers:', self.get_all_headers())
        print('count:', str(self.count))
        print('total:', str(self.total))
        print('resources:', self.resources)


class UpdateResourceResponse(LogResponse):
    """ The response of the update resource API from log.

    :type resp: dict
    :param resp: UpdateResourceResponse HTTP response body

    :type header: dict
    :param header: UpdateResourceResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateResourceResponse:')
        print('headers:', self.get_all_headers())


class GetResourceResponse(LogResponse):
    """ The response of the get resource API from log.

    :type resp: dict
    :param resp: GetResourceResponse HTTP response body

    :type header: dict
    :param header: GetResourceResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)
        self.resource = Resource.from_dict(resp)

    def get_resource(self):
        """Get resource from the response

        :return: instance of Resource
        """
        return self.resource

    def log_print(self):
        print('GetResourceResponse:')
        print('headers:', self.get_all_headers())
        print('resource: ', self.resource)


class CreateRecordResponse(LogResponse):
    """ The response of the create record API from log.

        :type resp: dict
        :param resp: CreateRecordResponse HTTP response body

        :type header: dict
        :param header: CreateRecordResponse HTTP response header
        """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateRecordResponse:')
        print('headers:', self.get_all_headers())


class DeleteRecordResponse(LogResponse):
    """ The response of the delete record API from log.

    :type resp: dict
    :param resp: DeleteRecordResponse HTTP response body

    :type header: dict
    :param header: DeleteRecordResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteRecordResponse:')
        print('headers:', self.get_all_headers())


class UpdateRecordResponse(LogResponse):
    """ The response of the update record API from log.

        :type resp: dict
        :param resp: UpdateRecordResponse HTTP response body

        :type header: dict
        :param header: UpdateRecordResponse HTTP response header
        """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateRecordResponse:')
        print('headers:', self.get_all_headers())


class UpsertRecordResponse(LogResponse):
    """ The response of the update or insert record API from log.

        :type resp: dict
        :param resp: UpsertRecordResponse HTTP response body

        :type header: dict
        :param header: UpsertRecordResponse HTTP response header
        """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpsertRecordResponse:')
        print('headers:', self.get_all_headers())


class GetRecordResponse(LogResponse):
    """ The response of the get record API from log.

    :type resp: dict
    :param resp: GetRecordResponse HTTP response body

    :type header: dict
    :param header: GetRecordResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)
        self.record = ResourceRecord.from_dict(resp)

    def get_record(self):
        """Get record from the response

        :return: instance of ResourceRecord
        """
        return self.record

    def log_print(self):
        print('GetRecordResponse:')
        print('headers:', self.get_all_headers())
        print('record:', self.record)


class ListRecordResponse(LogResponse):
    """ The response of the list_records API from log.

    :type header: dict
    :param header: ListRecordResponse HTTP response header

    :type resp: dict
    :param resp: ListRecordResponse HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total = int(resp["total"])
        self.records = [ResourceRecord.from_dict(record) for record in Util.convert_unicode_to_str(resp.get("items", []))]

    def get_records(self):
        """Get all the records from the response

        :return: list, instances of ResourceRecord
        """
        return self.records

    def get_count(self):
        """ Get count of records from the response

        :return: int, the number of count records from the response
        """
        return self.count

    def get_total(self):
        """Get total count of records from the response

        :return: int, the number of total records from the response
        """
        return self.total

    def log_print(self):
        print('ListRecordsResponse:')
        print('headers:', self.get_all_headers())
        print('count:', str(self.count))
        print('total:', str(self.total))
        print('records:', self.records)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logresponse import LogResponse
from .histogram import Histogram
from .util import Util


class GetHistogramsResponse(LogResponse):
    """ The response of the GetHistograms API from log.

    :type resp: dict
    :param resp: GetHistogramsResponse HTTP response body
    
    :type header: dict
    :param header: GetHistogramsResponse HTTP response header
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.progress = Util.h_v_t(header, 'x-log-progress')
        self.count = 0  # header['x-log-count']
        self.histograms = []

        for data in resp:
            status = Histogram(data['from'], data['to'], data['count'], data['progress'])
            self.histograms.append(status)
            self.count += data['count']

    def is_completed(self):
        """ Check if the histogram is completed
        
        :return: bool, true if this histogram is completed
        """
        return self.progress == 'Complete'

    def get_total_count(self):
        """ Get total logs' count that current query hits
        
        :return: int, total logs' count that current query hits
        """
        return self.count

    def get_histograms(self):
        """ Get histograms on the requested time range: [from, to)
        
        :return: Histogram list, histograms on the requested time range: [from, to)
        """
        return self.histograms

    def log_print(self):
        print('GetHistogramsResponse:')
        print('headers:', self.get_all_headers())
        print('progress:', self.progress)
        print('count:', self.count)
        print('\nhistograms class:\n')
        for data in self.histograms:
            data.log_print()
            print('\n')



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

from .logrequest import LogRequest


class ListLogstoresRequest(LogRequest):
    """ The request used to list log store from log.
    
    :type project: string
    :param project: project name
    """

    def __init__(self, project=None):
        LogRequest.__init__(self, project)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


from .logresponse import LogResponse
from .util import Util

__all__ = [
    'CreateProjectResponse', 'DeleteProjectResponse',
    'GetProjectResponse', 'ListProjectResponse',
    'GetProjectTagsResponse',
]


class CreateProjectResponse(LogResponse):
    """
    Response of create_project
    """
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateProjectResponse:')
        print('headers:', self.get_all_headers())


class DeleteProjectResponse(LogResponse):
    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteProjectResponse:')
        print('headers:', self.get_all_headers())


class GetProjectResponse(LogResponse):
    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.status = resp['status']
        self.description = resp['description']
        self.projectName = resp['projectName']
        self.region = resp['region']
        self.owner = resp['owner']
        self.createTime = resp['createTime']
        self.lastModifyTime = resp['lastModifyTime']

    def get_owner(self):
        return self.owner

    def get_status(self):
        return self.status

    def get_description(self):
        return self.description

    def get_projectname(self):
        return self.projectName

    def get_region(self):
        return self.region

    def get_create_time(self):
        return self.createTime

    def get_last_modify_time(self):
        return self.lastModifyTime

    def log_print(self):
        print('GetProjectResponse:')
        print('headers:', self.get_all_headers())
        print('owner:' + self.get_owner())
        print('project:' + self.get_projectname())
        print('description:' + self.get_description())
        print('region:' + self.get_region())
        print('status:' + self.get_status())
        print('create_time:' + self.get_create_time())
        print('last_modify_time:' + self.get_last_modify_time())


class ListProjectResponse(LogResponse):
    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp['count'])
        self.total = int(resp['total'])
        self.projects = Util.convert_unicode_to_str(resp.get("projects", []))

    def get_count(self):
        return self.count

    def get_total(self):
        return self.total

    def get_projects(self):
        return self.projects

    def log_print(self):
        print('ListProjectResponse:')
        print('headers:', self.get_all_headers())
        print('count:', self.count)
        print('total:', self.total)
        print('projects:', self.get_projects())

    def merge(self, response):
        if not isinstance(response, ListProjectResponse):
            raise ValueError("passed response is not a ListProjectResponse: " + str(type(response)))

        self.count += response.get_count()
        self.total = response.get_total() # use the latest total count
        self.projects.extend(response.get_projects())

        # update body
        self.body = {
            'count': self.count,
            'total': self.total,
            'projects': self.projects
        }

        return self


class GetProjectTagsResponse(LogResponse):
    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)
        self._next_token = resp["nextToken"]
        self._tags = {tag["tagKey"]: tag["tagValue"] for tag in resp["tagResources"]}

    def get_tags(self):
        return self._tags

    @property
    def next_token(self):
        return self._next_token



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


import json

from .. import LogItem
from ..util import parse_timestamp


class DocLogItemConverter(object):
    INDEX_FIELD = "_index"

    TYPE_FIELD = "_type"

    ID_FIELD = "_id"

    SOURCE_FIELD = "_source"

    TAG_PREFIX = "__tag__"

    @classmethod
    def to_log_item(cls, doc, time_reference):
        log_item = LogItem()
        cls._add_index(doc, log_item)
        cls._add_type(doc, log_item)
        cls._add_id(doc, log_item)
        cls._add_source(doc, log_item, time_reference)
        return log_item

    @classmethod
    def get_index(cls, doc):
        return doc[cls.INDEX_FIELD]

    @classmethod
    def _add_index(cls, doc, log_item):
        if cls.INDEX_FIELD not in doc:
            return
        log_item.push_back("__tag__:__es_index__", doc[cls.INDEX_FIELD])

    @classmethod
    def _add_type(cls, doc, log_item):
        if cls.TYPE_FIELD not in doc:
            return
        log_item.push_back("__tag__:__es_type__", doc[cls.TYPE_FIELD])

    @classmethod
    def _add_id(cls, doc, log_item):
        if cls.ID_FIELD not in doc:
            return
        log_item.push_back(cls.ID_FIELD, str(doc[cls.ID_FIELD]))

    @classmethod
    def _add_source(cls, doc, log_item, time_reference):
        if cls.SOURCE_FIELD not in doc:
            return
        for k, v in doc[cls.SOURCE_FIELD].items():
            if k == time_reference:
                log_item.set_time(parse_timestamp(v))
            if isinstance(v, dict):
                log_item.push_back(k, json.dumps(v))
            else:
                log_item.push_back(k, str(v))



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.


class QueriedLog(object):
    """ The QueriedLog is a log of the GetLogsResponse which obtained from the log.

    :type timestamp: int
    :param timestamp: log timestamp
    
    :type source: string
    :param source: log source
    
    :type contents: dict
    :param contents: log contents, content many key/value pair
    """

    def __init__(self, timestamp, source, contents):
        self.timestamp = int(timestamp)
        self.source = source
        self.contents = contents

    def get_time(self):
        """ Get log time
        
        :return: int, log time
        """
        return self.timestamp

    def get_source(self):
        """ Get log source
        
        :return: string, log source
        """
        return self.source

    def get_contents(self):
        """ Get log contents
        
        :return: dict, log contents
        """
        return self.contents

    def log_print(self):
        print('QueriedLog:')
        print('time:', self.get_time())
        print('source:', self.get_source())
        print('contents:', self.get_contents())



import copy
import inspect
import logging
import os
import sys

from .config_parser import ConfigParser
from .exceptions import SettingError
from .etl_util import process_event, u

logger = logging.getLogger(__name__)


class Runner(object):
    def __init__(self, config_path):
        config_path = u(config_path)

        if not inspect.ismodule(config_path):
            basedir = os.path.dirname(os.path.abspath(config_path))
            module_name = os.path.basename(config_path[:-3])
            if basedir not in sys.path:
                sys.path.insert(0, basedir)
            try:
                md = __import__(module_name)
            except ImportError as ex:
                logger.error(u"Cannot import config path: {0}".format(config_path))
                raise SettingError(ex, u'Cannot import the config "{0}"'.format(config_path))
        else:
            md = config_path

        logger.info(u"runner: passed module {0} from config file {1}".format(md, config_path))

        parsed_fn = ConfigParser(md).parse()
        logger.info(u"runner: passed fn list: {0}".format(parsed_fn))

        self.fn_list = [fn for no, fn in parsed_fn]

    def __call__(self, event):
        return process_event(event, self.fn_list)




# -*- coding: utf-8 -*-


import logging
import time
from threading import Thread

from .consumer_client import ConsumerClient

from .heart_beat import ConsumerHeatBeat
from .shard_worker import ShardConsumerWorker
from concurrent.futures import ThreadPoolExecutor


class ConsumerWorkerLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        consumer_worker = self.extra['consumer_worker']  # type: ConsumerWorker
        consumer_option = consumer_worker.option
        _id = '/'.join([
            consumer_option.project, consumer_option.logstore,
            consumer_option.consumer_group_name, consumer_option.consumer_name
        ])
        return "[{0}]{1}".format(_id, msg), kwargs


class ConsumerWorker(Thread):
    def __init__(self, make_processor, consumer_option, args=None,
                 kwargs=None):
        super(ConsumerWorker, self).__init__()
        self.make_processor = make_processor
        self.process_args = args or ()
        self.process_kwargs = kwargs or {}
        self.option = consumer_option
        self.consumer_client = \
            ConsumerClient(consumer_option.endpoint, consumer_option.accessKeyId, consumer_option.accessKey,
                           consumer_option.project, consumer_option.logstore, consumer_option.consumer_group_name,
                           consumer_option.consumer_name, consumer_option.securityToken,
                           credentials_refresher=consumer_option.credentials_refresher)
        self.shut_down_flag = False
        self.logger = ConsumerWorkerLoggerAdapter(
            logging.getLogger(__name__), {"consumer_worker": self})
        self.shard_consumers = {}

        self.last_owned_consumer_finish_time = 0

        self.consumer_client.create_consumer_group(consumer_option.consumer_group_time_out, consumer_option.in_order)
        self.heart_beat = ConsumerHeatBeat(self.consumer_client, consumer_option.heartbeat_interval,
                                           consumer_option.consumer_group_time_out)

        if consumer_option.shared_executor is not None:
            self.own_executor = False
            self._executor = consumer_option.shared_executor
        else:
            self.own_executor = True
            self._executor = ThreadPoolExecutor(max_workers=consumer_option.worker_pool_size)

    @property
    def executor(self):
        return self._executor

    def _need_stop(self):
        """
        check if need to stop:
        1. end_cursor has been hit and there's no more shard assinged (wait for heatbeat_interval * 3)
        :return:
        """
        if not self.option.cursor_end_time:
            return False

        all_finish = True
        for shard, consumer in self.shard_consumers.items():
            if consumer.is_shutdown():
                continue

            # has not yet do any successful fetch yet or get some data
            if consumer.last_success_fetch_time == 0 or consumer.last_fetch_count > 0:
                return False

        # init self.last_owned_consumer_finish_time if it's None
        if all_finish and self.last_owned_consumer_finish_time == 0:
            self.last_owned_consumer_finish_time = time.time()

        if abs(time.time() - self.last_owned_consumer_finish_time) >= \
                self.option.consumer_group_time_out + self.option.heartbeat_interval:
            return True

        return False

    def run(self):
        self.logger.info('consumer worker "{0}" start '.format(self.option.consumer_name))
        self.heart_beat.start()

        while not self.shut_down_flag:
            held_shards = self.heart_beat.get_held_shards()

            last_fetch_time = time.time()
            for shard in held_shards:
                if self.shut_down_flag:
                    break

                shard_consumer = self._get_shard_consumer(shard)
                if shard_consumer is None:  # error when init consumer. shutdown directly
                    self.shutdown()
                    break

                shard_consumer.consume()

            self.clean_shard_consumer(held_shards)

            if self._need_stop():
                self.logger.info("all owned shards complete the tasks, owned shards: {0}".format(self.shard_consumers))
                self.shutdown()

            time_to_sleep = self.option.data_fetch_interval - (time.time() - last_fetch_time)
            while time_to_sleep > 0 and not self.shut_down_flag:
                time.sleep(min(time_to_sleep, 1))
                time_to_sleep = self.option.data_fetch_interval - (time.time() - last_fetch_time)

        # # stopping worker, need to cleanup all existing shard consumer
        self.logger.info('consumer worker "{0}" try to cleanup consumers'.format(self.option.consumer_name))
        self.shutdown_and_wait()

        if self.own_executor:
            self.logger.info('consumer worker "{0}" try to shutdown executors'.format(self.option.consumer_name))
            self._executor.shutdown()
            self.logger.info('consumer worker "{0}" stopped'.format(self.option.consumer_name))
        else:
            self.logger.info('executor is shared, consumer worker "{0}" stopped'.format(self.option.consumer_name))

    def start(self, join=False):
        """
        when calling with join=True, must call it in main thread, or else, the Keyboard Interrupt won't be caputured.
        :param join: default False, if hold on until the worker is stopped by Ctrl+C or other reasons.
        :return:
        """
        Thread.start(self)

        if join:
            try:
                while self.is_alive():
                    self.join(timeout=60)
                self.logger.info("worker {0} exit unexpected, try to shutdown it".format(self.option.consumer_name))
                self.shutdown()
            except KeyboardInterrupt:
                self.logger.info("*** try to exit **** ")
                self.shutdown()

    def shutdown_and_wait(self):
        while True:
            time.sleep(0.5)
            for shard, consumer in self.shard_consumers.items():
                if not consumer.is_shutdown():
                    consumer.shut_down()
                    break  # there's live consumer, no need to check, loop to next
            else:
                break   # all are shutdown, exit look

        self.shard_consumers.clear()

    def clean_shard_consumer(self, owned_shards):
        remove_shards = []
        # remove the shards that's not assigned by server
        for shard, consumer in self.shard_consumers.items():
            if shard not in owned_shards:
                self.logger.info('Try to call shut down for unassigned consumer shard: ' + str(shard))
                consumer.shut_down()
                self.logger.info('Complete call shut down for unassigned consumer shard: ' + str(shard))
            if consumer.is_shutdown():
                self.heart_beat.remove_heart_shard(shard)
                remove_shards.append(shard)
                self.logger.info('Remove an unassigned consumer shard:' + str(shard))

        for shard in remove_shards:
            self.shard_consumers.pop(shard)

    def shutdown(self):
        self.shut_down_flag = True
        self.heart_beat.shutdown()
        self.logger.info('get stop signal, start to stop consumer worker "{0}"'.format(self.option.consumer_name))

    def _get_shard_consumer(self, shard_id):
        consumer = self.shard_consumers.get(shard_id, None)
        if consumer is not None:
            return consumer

        try:
            processer = self.make_processor(*self.process_args, **self.process_kwargs)
        except Exception as ex:
            self.logger.error("fail to init processor {0} with parameters {1}, {2}, detail: {3}".format(
                self.make_processor, self.process_args, self.process_kwargs, ex, exc_info=True))
            return None

        consumer = ShardConsumerWorker(self.consumer_client, shard_id, self.option.consumer_name,
                                       processer,
                                       self.option.cursor_position, self.option.cursor_start_time,
                                       executor=self._executor,
                                       cursor_end_time=self.option.cursor_end_time,
                                       max_fetch_log_group_size=self.option.max_fetch_log_group_size)
        self.shard_consumers[shard_id] = consumer
        return consumer



# -*- coding: utf-8 -×-


class FetchedLogGroup(object):

    def __init__(self, shard_id, log_group_list, end_cursor):
        self._shard_id = shard_id
        self._fetched_log_group_list = log_group_list
        self._end_cursor = end_cursor

    @property
    def shard_id(self):
        return self._shard_id

    @property
    def fetched_log_group_list(self):
        return self._fetched_log_group_list

    @property
    def end_cursor(self):
        return self._end_cursor

    @property
    def log_group_size(self):
        return len(self._fetched_log_group_list.LogGroups)



# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.
import logging
from .logexception import LogException

logger = logging.getLogger(__name__)

class ExternalStoreConfigBase(object):
    def __init__(self, externalStoreName, storeType):
        pass

    def to_json(self):
        pass

    @staticmethod
    def from_json(json_value):
        storeType = json_value["storeType"]
        if storeType == 'rds-vpc':
            return ExternalStoreConfig.from_json(json_value)
        elif storeType == 'oss':
            return ExternalStoreOssConfig.from_json(json_value)
        elif storeType == 'csv':
            return ExternalStoreCsvConfig.from_json(json_value)
        else:
            raise LogException("Unknown storeType", "please contact support")

    def log_print(self):
        pass


class ExternalStoreConfig(ExternalStoreConfigBase):
    """

    """

    def __init__(self, externalStoreName, region, storeType, vpcId, instanceId, host, port, username, password,
                 database, table):
        self.externalStoreName = externalStoreName
        self.storeType = storeType
        self.vpcId = vpcId
        self.instanceId = instanceId
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.table = table
        self.region = region

    def to_json(self):
        json_value = {}
        json_value["externalStoreName"] = self.externalStoreName
        json_value["storeType"] = self.storeType
        param = {}
        param["vpc-id"] = self.vpcId
        param["instance-id"] = self.instanceId
        param["host"] = self.host
        param["port"] = self.port
        param["username"] = self.username
        param["password"] = self.password
        param["db"] = self.database
        param["table"] = self.table
        param["region"] = self.region
        json_value["parameter"] = param
        return json_value

    @staticmethod
    def from_json(json_value):
        storeType = json_value["storeType"]
        externalStoreName = json_value["externalStoreName"]
        vpcId = json_value["parameter"].get("vpc-id", "")
        instanceId = json_value["parameter"].get("instance-id", "")
        host = json_value["parameter"].get("host", "")
        port = json_value["parameter"]["port"]
        username = json_value["parameter"]["username"]
        password = json_value["parameter"].get("password", "")
        database = json_value["parameter"]["db"]
        table = json_value["parameter"]["table"]
        region = json_value["parameter"]["region"]

        return ExternalStoreConfig(externalStoreName, region, storeType, vpcId, instanceId,
                                   host, port, username, password, database, table)

    def log_print(self):
        print("region", self.region)
        print ("storeType", self.storeType)
        print("vpcId", self.vpcId)
        print("instanceId", self.instanceId)
        print("host", self.host)
        print("port", self.port)
        print("username", self.username)
        print("password", self.password)
        print("db", self.database)
        print("table", self.table)


class ExternalStoreOssConfig(ExternalStoreConfigBase):
    """

    """

    def __init__(self, externalStoreName, objects, columns, endpoint, bucket, accessid, accesskey):
        self.externalStoreName = externalStoreName
        self.storeType = 'oss'
        self.objects = objects
        self.columns = columns
        self.bucket = bucket
        self.accessid = accessid
        self.accesskey = accesskey
        self.endpoint = endpoint

    def to_json(self):
        json_value = {}
        json_value["externalStoreName"] = self.externalStoreName
        json_value["storeType"] = self.storeType
        param = {}
        param["endpoint"] = self.endpoint
        param["bucket"] = self.bucket
        param["accessid"] = self.accessid
        param["accesskey"] = self.accesskey
        param["columns"] = self.columns
        param["objects"] = self.objects
        json_value["parameter"] = param
        return json_value

    @staticmethod
    def from_json(json_value):
        externalStoreName = json_value["externalStoreName"]
        storeType = json_value["storeType"]
        endpoint = json_value["parameter"]["endpoint"]
        bucket = json_value["parameter"]["bucket"]
        accessid = json_value["parameter"]["accessid"]
        accesskey = json_value["parameter"]["accesskey"]
        columns = json_value["parameter"]["columns"]
        objects = json_value["parameter"]["objects"]

        return ExternalStoreOssConfig(externalStoreName, objects, columns,
                                      endpoint, bucket, accessid, accesskey)

    def log_print(self):
        print("externalStoreName", self.externalStoreName)
        print ("storeType", self.storeType)
        print("endpoint", self.endpoint)
        print("bucket", self.bucket)
        print("accessid", self.accessid)
        print("accesskey", self.accesskey)
        print("columns", self.columns)
        print("objects", self.objects)

class ExternalStoreCsvConfig(ExternalStoreConfigBase):
    def __init__(self, externalStoreName, externalStoreCsvFile, columns, resp = False):
        import zlib
        import base64
        self.externalStoreName = externalStoreName
        self.columns = columns
        self.objects = [externalStoreCsvFile,]
        self.storeType = 'csv'
        if not resp:
            externalStoreCsvFh = open(externalStoreCsvFile, 'rb')
            externalStoreCsv = externalStoreCsvFh.read()
            externalStoreCsvFh.close()
            self.externalStoreCsvSize = len(externalStoreCsv)
            maxOriSize = 50 * 1000 * 1000
            maxSize = 10 * 1000 * 1000  - 10 * 1000
            if self.externalStoreCsvSize > maxOriSize:
                # too large
                raise LogException("ExternalStoreCsvConfig", "csv file size too large, max " + str(maxOriSize))
            externalStoreCsvCompressed = zlib.compress(externalStoreCsv)
            self.externalStoreCsvCompressedBase64 = base64.standard_b64encode(externalStoreCsvCompressed).decode()
            if len(self.externalStoreCsvCompressedBase64) > maxSize:
                raise LogException("ExternalStoreCsvConfig",
                                   "size of csv file after compressed too large, max " + str(maxSize)
                                   + "now size: " + str(len(self.externalStoreCsvCompressedBase64)))

    def to_json(self):
        json_value = {}
        json_value["externalStoreName"] = self.externalStoreName
        json_value["storeType"] = self.storeType
        param = {}
        param["columns"] = self.columns
        param["objects"] = self.objects
        param["externalStoreCsv"] = self.externalStoreCsvCompressedBase64
        param["externalStoreCsvSize"] = self.externalStoreCsvSize
        json_value["parameter"] = param
        return json_value

    @staticmethod
    def from_json(json_value):
        externalStoreName = json_value["externalStoreName"]
        storeType = json_value["storeType"]
        endpoint = json_value["parameter"]["endpoint"]
        columns = json_value["parameter"]["columns"]
        objects = json_value["parameter"]["objects"]

        return ExternalStoreCsvConfig(externalStoreName, objects, columns, True)

    def log_print(self):
        print("externalStoreName", self.externalStoreName)
        print("storeType", self.storeType)
        print("columns", self.columns)
        #print("objects", self.objects)





from .logresponse import LogResponse
from .util import Util

__all__ = ['CreateEtlResponse', 'DeleteEtlResponse', 'GetEtlResponse', 'UpdateEtlResponse',
           'ListEtlsResponse', 'StopEtlResponse', 'StartEtlResponse']


class CreateEtlResponse(LogResponse):
    """ The response of the create_etl API from log.
    :type resp: dict
    :param resp: CreateEtlResponse HTTP response body
    :type header: dict
    :param header: CreateEtlResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('CreateEtlResponse:')
        print('headers:', self.get_all_headers())


class DeleteEtlResponse(LogResponse):
    """ The response of the delete_etl API from log.
    :type resp: dict
    :param resp: DeleteEtlResponse HTTP response body
    :type header: dict
    :param header: DeleteEtlResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('DeleteEtlResponse:')
        print('headers:', self.get_all_headers())


class GetEtlResponse(LogResponse):
    """ The response of the get_etl API from log.
    :type resp: dict
    :param resp: GetEtlResponse HTTP response body
    :type header: dict
    :param header: GetEtlResponse HTTP response header
    """

    def __init__(self, header, resp):
        LogResponse.__init__(self, header, resp)
        self.etl_config = resp

    def get_etl(self):
        """Get etl from the response
        :return: dict, etl
        """
        return self.etl_config

    def log_print(self):
        print('GetEtlResponse:')
        print('headers:', self.get_all_headers())
        print('etl: ', self.etl_config)


class UpdateEtlResponse(LogResponse):
    """ The response of the update_etl API from log.
    :type resp: dict
    :param resp: UpdateEtlResponse HTTP response header
    :type header: dict
    :param header: UpdateEtlResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('UpdateEtlResponse:')
        print('headers:', self.get_all_headers())


class StartEtlResponse(LogResponse):
    """ The response of the start_etl API from log.
    :type resp: dict
    :param resp: StartEtlResponse HTTP response body
    :type header: dict
    :param header: StartEtlResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('StartEtlResponse:')
        print('headers:', self.get_all_headers())


class StopEtlResponse(LogResponse):
    """ The response of the stop_etl API from log.
    :type resp: dict
    :param resp: StopEtlResponse HTTP response body
    :type header: dict
    :param header: StopEtlResponse HTTP response header
    """

    def __init__(self, header, resp=''):
        LogResponse.__init__(self, header, resp)

    def log_print(self):
        print('StopEtlResponse:')
        print('headers:', self.get_all_headers())


class ListEtlsResponse(LogResponse):
    """ The response of the list_etls API from log.
    :type header: dict
    :param header: ListEtlsResponse HTTP response header
    :type resp: dict
    :param resp: ListEtlsResponse HTTP response body
    """

    def __init__(self, resp, header):
        LogResponse.__init__(self, header, resp)
        self.count = int(resp["count"])
        self.total = int(resp["total"])
        self.etls = Util.convert_unicode_to_str(resp.get("results", []))

    def get_etls(self):
        """Get all the etls from the response
        :return: list, all etls
        """
        return self.etls

    def get_count(self):
        """ Get count of etls from the response
        :return: int, the number of count etls from the response
        """
        return self.count

    def get_total(self):
        """Get total count of etls from the response
        :return: int, the number of total etls from the response
        """
        return self.total

    def log_print(self):
        print('ListEtlResponse:')
        print('headers:', self.get_all_headers())
        print('count:', str(self.count))
        print('total:', str(self.total))
        print('etls:', self.etls)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) Alibaba Cloud Computing
# All rights reserved.

import time


class IndexJsonKeyConfig(object):
    """ The index config of a special json type key

    :type index_all: bool
    :param index_all: True if all string value in the json key should be indexed

    :type max_depth: int
    :param max_depth: if index_all is true, only if the json value depth <= max_depth will be index

    :type alias : string
    :param alias : alias name for index key
    """

    def __init__(self, index_all=True, max_depth=-1, alias=None):
        self.index_all = index_all
        self.max_depth = max_depth
        self.alias = alias
        self.json_keys = {}

    """
    Inner key config in json, if the json value is : 
    json_key : {
        "map_1" : {
            "k_1" : "v_1",
            "k_2" : 100
        },
        "k_3" : 200.0
    }

    :type key_name : string
    :param key_name : key name , e.g  "map_1.k_1",  "k_3"

    :type index_type: string
    :param index_type: one of ['text', 'long', 'double']

    :type doc_value : bool
    :param doc_value : if save doc value

    :type alias : string
    :param alias : alias name for index key
    """

    def add_key(self, key_name, key_type, doc_value=False, alias=None):
        if key_type != 'text' and key_type != 'long' and key_type != 'double':
            return
        self.json_keys[key_name] = {}
        self.json_keys[key_name]["type"] = key_type
        self.json_keys[key_name]["doc_value"] = doc_value
        if alias is not None:
            self.json_keys[key_name]["alias"] = alias

    def to_json(self, json_value):
        json_value["index_all"] = self.index_all
        json_value["max_depth"] = self.max_depth
        json_value["json_keys"] = self.json_keys

    def from_json(self, json_value):
        self.index_all = json_value.get("index_all", True)
        self.max_depth = json_value.get("max_depth", -1)
        self.alias = None
        self.json_keys = {}
        if "alias" in json_value:
            self.alias = json_value["alias"]
        if "json_keys" in json_value:
            self.json_keys = json_value["json_keys"]


class IndexKeyConfig(object):
    """ The index config of a special log key

    :type token_list: string list
    :param token_list: the token config list, e.g ["," , "\t" , "\n" , " " , ";"]

    :type case_sensitive: bool
    :param case_sensitive: True if the value in the log keys is case sensitive, False other wise

    :type index_type: string
    :param index_type: one of ['text', 'long', 'double', 'json']

    :type doc_value: bool
    :param doc_value: True if enable doc_value, used for fast sql execution

    :type alias : string
    :param alias : alias name for index key

    :type json_key_config : IndexJsonKeyConfig
    :param json_key_config : configs for "json" type

    :type chinese: bool
    :param chinese: enable Chinese words segmentation

    """

    def __init__(self, token_list=None, case_sensitive=False, index_type='text', doc_value=False, alias=None,
                 json_key_config=None, chinese=None):
        if token_list is None:
            token_list = []
        self.token_list = token_list
        self.case_sensitive = case_sensitive
        self.index_type = index_type
        self.doc_value = doc_value
        self.alias = alias
        self.json_key_config = json_key_config
        self.chn = chinese

    def set_json_key_config(self, json_key_config):
        self.json_key_config = json_key_config

    def get_json_key_config(self):
        return self.json_key_config

    def to_json(self):
        json_value = {}
        if self.index_type != "" and self.index_type is not None:
            json_value['type'] = self.index_type
        if self.index_type == 'text' or self.index_type == 'json':
            json_value["token"] = self.token_list
            json_value["caseSensitive"] = bool(self.case_sensitive)
        if self.alias is not None:
            json_value['alias'] = self.alias
        json_value["doc_value"] = bool(self.doc_value)

        if self.chn is not None:
            json_value['chn'] = self.chn

        if self.index_type == "json":
            self.json_key_config.to_json(json_value)

        return json_value

    def from_json(self, json_value):
        self.index_type = 'text'
        if 'type' in json_value:
            self.index_type = json_value['type']
        if self.index_type in ('text', 'json'):
            self.token_list = json_value["token"]
            self.chn = None
            if "chn" in json_value:
                self.chn = bool(json_value["chn"])
        self.case_sensitive = bool(json_value.get("caseSensitive", False))
        if 'doc_value' in json_value:
            self.doc_value = bool(json_value["doc_value"])
        if 'alias' in json_value:
            self.alias = json_value['alias']
        if self.index_type == 'json':
            self.json_key_config = IndexJsonKeyConfig()
            self.json_key_config.from_json(json_value)


class IndexLineConfig(object):
    """ The index config of the log line

    :type token_list: string list
    :param token_list: the token config list, e.g ["," , "\t" , "\n" , " " , ";"]

    :type case_sensitive: bool
    :param case_sensitive: True if the value in the log keys is case sensitive, False other wise

    :type include_keys: string list
    :param include_keys: deprecated, will be removed in future version.

    :type exclude_keys: string list
    :param exclude_keys: deprecated, will be removed in future version.

    :type chinese: bool
    :param chinese: enable Chinese words segmentation

    """

    def __init__(self, token_list=None, case_sensitive=False, include_keys=None, exclude_keys=None, chinese=None):
        if token_list is None:
            token_list = []
        self.token_list = token_list
        self.case_sensitive = case_sensitive
        self.chn = bool(chinese)

    def to_json(self):
        json_value = {"token": self.token_list, "caseSensitive": bool(self.case_sensitive)}
        if self.chn is not None:
            json_value["chn"] = bool(self.chn)

        return json_value

    def from_json(self, json_value):
        self.token_list = json_value["token"]
        self.case_sensitive = bool(json_value.get("caseSensitive", False))
        self.chn = bool(json_value["caseSensitive"]) if "caseSensitive" in json_value else None


class IndexConfig(object):
    """The index config of a logstore

    :type ttl: int
    :param ttl: this parameter is deprecated, the ttl is same as logstore's ttl

    :type line_config: IndexLineConfig
    :param line_config: the index config of the whole log line

    :type key_config_list: dict
    :param key_config_list: dict (string => IndexKeyConfig), the index key configs of the keys

    :type all_keys_config: IndexKeyConfig
    :param all_keys_config: the key config of all keys, the new create logstore should never user this param, it only used to compatible with old config

    :type log_reduce: bool
    :param log_reduce: if to enable logreduce

    """

    def __init__(self, ttl=1, line_config=None, key_config_list=None, all_keys_config=None, log_reduce=None):
        if key_config_list is None:
            key_config_list = {}
        self.ttl = ttl
        self.all_keys_config = all_keys_config
        self.line_config = line_config
        self.key_config_list = key_config_list
        self.modify_time = int(time.time())
        self.log_reduce = log_reduce
        self.log_reduce_black_list = None
        self.log_reduce_white_list = None
        self.docvalue_max_text_len = 0

    '''
    :type max_len : int 
    :param max_len : the max len  of the docvalue
    '''

    def set_docvalue_max_text_len(self, max_len):
        self.docvalue_max_text_len = max_len

    '''
    :type black_list : list 
    :param black_list : the list of black list keys for log reduce, e.g black_list = ["key_1", "key_2"] 
    '''

    def set_log_reduce_black_list(self, black_list):
        self.log_reduce_black_list = black_list

    '''
    :type white_list : list 
    :param white_list : the list of white list keys for log reduce, e.g white_list = ["key_1", "key_2"] 
    '''

    def set_log_reduce_white_list(self, white_list):
        self.log_reduce_white_list = white_list

    def to_json(self):
        json_value = {}
        if self.line_config is not None:
            json_value["line"] = self.line_config.to_json()
        if len(self.key_config_list) != 0:
            json_value["keys"] = dict((key, value.to_json()) for key, value in self.key_config_list.items())

        if self.all_keys_config is not None:
            json_value["all_keys"] = self.all_keys_config.to_json()

        if self.log_reduce is not None:
            json_value["log_reduce"] = self.log_reduce

        if self.docvalue_max_text_len > 0:
            json_value["max_text_len"] = self.docvalue_max_text_len

        if self.log_reduce_white_list != None:
            json_value["log_reduce_white_list"] = self.log_reduce_white_list

        elif self.log_reduce_black_list != None:
            json_value["log_reduce_black_list"] = self.log_reduce_black_list

        return json_value

    def from_json(self, json_value):
        self.ttl = json_value.get("ttl", 0)
        if "all_keys" in json_value:
            self.all_keys_config = IndexKeyConfig()
            self.all_keys_config.from_json(json_value["all_keys"])
        if "line" in json_value:
            self.line_config = IndexLineConfig()
            self.line_config.from_json(json_value["line"])
        if "keys" in json_value:
            self.key_config_list = {}
            key_configs = json_value["keys"]
            for key, value in key_configs.items():
                key_config = IndexKeyConfig()
                key_config.from_json(value)
                self.key_config_list[key] = key_config
        self.log_reduce = json_value.get('log_reduce', None)
        self.log_reduce_white_list = json_value.get('log_reduce_white_list', None)
        self.log_reduce_black_list = json_value.get('log_reduce_black_list', None)

        self.docvalue_max_text_len = json_value.get('max_text_len', 0)

        self.modify_time = json_value.get("lastModifyTime", int(time.time()))



# -*- coding: utf-8 -*-

import copy
import logging
import time

from .checkpoint_tracker import ConsumerCheckpointTracker

from .config import ConsumerStatus
from .fetched_log_group import FetchedLogGroup
from .tasks import ProcessTaskResult, InitTaskResult, FetchTaskResult, TaskResult
from .tasks import consumer_fetch_task, consumer_initialize_task, \
    consumer_process_task, consumer_shutdown_task
from .exceptions import ClientWorkerException
import six


class ShardConsumerWorkerLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        shard_consumer_worker = self.extra[
            'shard_consumer_worker']  # type: ShardConsumerWorker
        consumer_client = shard_consumer_worker.log_client
        _id = '/'.join([
            consumer_client.mproject, consumer_client.mlogstore,
            consumer_client.mconsumer_group, consumer_client.mconsumer,
            str(shard_consumer_worker.shard_id)
        ])
        return "[{0}]{1}".format(_id, msg), kwargs


class ShardConsumerWorker(object):
    def __init__(self, log_client, shard_id, consumer_name, processor, cursor_position, cursor_start_time,
                 max_fetch_log_group_size=1000, executor=None, cursor_end_time=None):
        self.log_client = log_client
        self.shard_id = shard_id
        self.consumer_name = consumer_name
        self.cursor_position = cursor_position
        self.cursor_start_time = cursor_start_time
        self.cursor_end_time = cursor_end_time or None
        self.processor = processor
        self.checkpoint_tracker = ConsumerCheckpointTracker(self.log_client, self.consumer_name,
                                                            self.shard_id)
        self.executor = executor
        self.max_fetch_log_group_size = max_fetch_log_group_size

        self.consumer_status = ConsumerStatus.INITIALIZING
        self.current_task_exist = False
        self.task_future = None
        self.fetch_data_future = None

        self.next_fetch_cursor = ''
        self.fetch_end_cursor = None

        self.shutdown = False
        self.last_fetch_log_group = None

        self.last_log_error_time = 0
        self.last_fetch_time = 0
        self.last_fetch_count = 0
        self.last_success_fetch_time = 0
        self.last_success_fetch_time_with_data = 0
        self.save_last_checkpoint = False

        self.logger = ShardConsumerWorkerLoggerAdapter(
            logging.getLogger(__name__), {"shard_consumer_worker": self})

    def consume(self):
        self.logger.debug('consumer start consuming')
        self.check_and_generate_next_task()
        if self.consumer_status == ConsumerStatus.PROCESSING and self.last_fetch_log_group is None:
            self.fetch_data()

    @staticmethod
    # get future (if failed return None)
    def get_task_result(task_future):
        if task_future is not None and task_future.done():
            try:
                return task_future.result()
            except Exception:
                import traceback
                traceback.print_exc()
        return None

    def fetch_data(self):
        # no task or it's done
        if self.fetch_data_future is None or self.fetch_data_future.done():
            task_result = self.get_task_result(self.fetch_data_future)

            # task is done, output results and get next_cursor
            if task_result is not None and task_result.get_exception() is None:
                assert isinstance(task_result, FetchTaskResult), \
                    ClientWorkerException("fetch result type is not as expected")

                self.last_success_fetch_time = time.time()

                self.last_fetch_log_group = FetchedLogGroup(self.shard_id, task_result.get_fetched_log_group_list(),
                                                            task_result.get_cursor())
                self.next_fetch_cursor = task_result.get_cursor()
                self.last_fetch_count = self.last_fetch_log_group.log_group_size
                if self.last_fetch_count > 0:
                    self.last_success_fetch_time_with_data = time.time()
                    self.save_last_checkpoint = False
                else:
                    if self.last_success_fetch_time_with_data != 0 and time.time() - self.last_success_fetch_time_with_data > 30 \
                            and not self.save_last_checkpoint:
                        self.checkpoint_tracker.flush_check_point()
                        self.save_last_checkpoint = True


            self._sample_log_error(task_result)

            # no task or task is done, create new task
            if task_result is None or task_result.get_exception() is None:
                # flag to indicate if it's done
                is_generate_fetch_task = True

                # throttling control, similar as Java's SDK
                if self.last_fetch_count < 100:
                    is_generate_fetch_task = (time.time() - self.last_fetch_time) > 0.5
                elif self.last_fetch_count < 500:
                    is_generate_fetch_task = (time.time() - self.last_fetch_time) > 0.2
                elif self.last_fetch_count < 1000:
                    is_generate_fetch_task = (time.time() - self.last_fetch_time) > 0.05

                if is_generate_fetch_task:
                    self.last_fetch_time = time.time()
                    self.fetch_data_future = \
                        self.executor.submit(consumer_fetch_task,
                                             self.log_client, self.shard_id, self.next_fetch_cursor,
                                             max_fetch_log_group_size=self.max_fetch_log_group_size,
                                             end_cursor=self.fetch_end_cursor)
                else:
                    self.fetch_data_future = None
            else:
                self.fetch_data_future = None

    def check_and_generate_next_task(self):
        """
        check if the previous task is done and proceed to fire another task
        :return:
        """
        # if self.task_future is None:
        #     # there's no any ongoing task
        #     self._update_status(False)
        #     self.generate_next_task()
        #     return

        if self.task_future is None or self.task_future.done():
            task_success = False

            task_result = self.get_task_result(self.task_future)
            self.task_future = None

            if task_result is not None and task_result.get_exception() is None:

                task_success = True
                if isinstance(task_result, InitTaskResult):
                    # maintain check points
                    assert self.consumer_status == ConsumerStatus.INITIALIZING, \
                        ClientWorkerException("get init task result, but status is: " + str(self.consumer_status))

                    init_result = task_result
                    self.next_fetch_cursor = init_result.get_cursor()
                    self.fetch_end_cursor = init_result.end_cursor
                    self.checkpoint_tracker.set_memory_check_point(self.next_fetch_cursor)
                    if init_result.is_cursor_persistent():
                        self.checkpoint_tracker.set_persistent_check_point(self.next_fetch_cursor)

                elif isinstance(task_result, ProcessTaskResult):
                    # maintain check points
                    process_task_result = task_result
                    roll_back_checkpoint = process_task_result.get_rollback_check_point()
                    if roll_back_checkpoint:
                        self.last_fetch_log_group = None
                        self.logger.info("user defined to roll-back check-point, cancel current fetching task")
                        self.cancel_current_fetch()
                        self.next_fetch_cursor = roll_back_checkpoint

            # log task status
            self._sample_log_error(task_result)

            # update status basing on task results
            self._update_status(task_success)

            #
            self._generate_next_task()

    def _generate_next_task(self):
        """
        submit consumer framework defined task
        :return:
        """
        if self.consumer_status == ConsumerStatus.INITIALIZING:
            self.current_task_exist = True
            self.task_future = self.executor.submit(consumer_initialize_task, self.processor, self.log_client,
                                                    self.shard_id, self.cursor_position, self.cursor_start_time, self.cursor_end_time)

        elif self.consumer_status == ConsumerStatus.PROCESSING:
            if self.last_fetch_log_group is not None:
                self.checkpoint_tracker.set_cursor(self.last_fetch_log_group.end_cursor)
                self.current_task_exist = True

                # must deep copy cause some revision will happen
                last_fetch_log_group = copy.deepcopy(self.last_fetch_log_group)
                self.last_fetch_log_group = None

                if self.last_fetch_count > 0:
                    self.task_future = self.executor.submit(consumer_process_task, self.processor,
                                                            last_fetch_log_group.fetched_log_group_list,
                                                            self.checkpoint_tracker)

        elif self.consumer_status == ConsumerStatus.SHUTTING_DOWN:
            self.current_task_exist = True
            self.logger.info("start to cancel fetch job")
            self.cancel_current_fetch()
            self.task_future = self.executor.submit(consumer_shutdown_task, self.processor, self.checkpoint_tracker)

    def cancel_current_fetch(self):
        if self.fetch_data_future is not None:
            self.fetch_data_future.cancel()
            self.logger.info('Cancel a fetch task, shard id: ' + str(self.shard_id))
            self.fetch_data_future = None

    def _sample_log_error(self, result):
        # record the time when error happens
        if not isinstance(result, TaskResult):
            return

        exc = result.get_exception()
        if exc is None:
            return

        current_time = time.time()
        if current_time - self.last_log_error_time <= 5:
            return

        self.logger.warning(exc, exc_info=result.exc_info)
        self.last_log_error_time = current_time

    def _update_status(self, task_succcess):
        if self.consumer_status == ConsumerStatus.SHUTTING_DOWN:
            # if no task or previous task suceeds, shutting-down -> shutdown complete
            if not self.current_task_exist or task_succcess:
                self.consumer_status = ConsumerStatus.SHUTDOWN_COMPLETE
        elif self.shutdown:
            # always set to shutting down (if flag is set)
            self.consumer_status = ConsumerStatus.SHUTTING_DOWN
        elif self.consumer_status == ConsumerStatus.INITIALIZING:
            # if initing and has task succeed, init -> processing
            if task_succcess:
                self.consumer_status = ConsumerStatus.PROCESSING

    def shut_down(self):
        """
        set shutting down flag, if not shutdown yet, complete the ongoing task(?)
        :return:
        """
        self.shutdown = True
        if not self.is_shutdown():
            self.check_and_generate_next_task()

    def is_shutdown(self):
        return self.consumer_status == ConsumerStatus.SHUTDOWN_COMPLETE
